#!/usr/bin/env perl
# 
# - sshd_autoban.pl - V0.75 -
# ** Changelog
#    V0.75 - Removed fifo usage, and using sockets instead
#          - Updated documentation
#    V0.7  - Add a function to clean old entries from hosts.deny
#          - Add syslogd configuration to the script documentation
#          - Some code cosmetics
#    V0.6  - Quick fix on a variable typo
#    V0.5  - Add an option to select iptables or hosts.deny ban type
# **
#
# This script is intended to automatically ban, in real time, script kiddies
# that attempt to log onto your sshd service.
# 
# To access sshd logs in real time, you have to set your syslog daemon to output
# sshd logs to a fifo file.
# ######################
# SSHD configuration:
# ######################
# First edit your sshd config and add/change those lines:
#
# SyslogFacility DEAMON
# LogLevel VERBOSE
#
# #######################
# Syslog-ng configuration:
# #######################
# Just add those lines to your /etc/syslog-ng/syslog-ng.conf:
#
# destination autoban { tcp("127.0.0.1" port(10200)); };
# destination ssh { file("/var/log/ssh.log"); };
# filter f_ssh { program("sshd"); };
# log { source(src); filter(f_ssh); destination(ssh); destination(autoban); };
#
# Restart syslog-ng, you're now ready to run the script.
# ##########################
# This script will use /etc/hosts.deny or iptables to ban IP addresses that do 5 unsuccessfull 
# connections attempts in less than 15 seconds, and will ban ip in /var/log/sshd_autoban.log
# 
# It sure needs root permissions to log and write to /etc/hosts.deny or use iptables.
# 
# Just modify values below to suit your needs:
# number of attempts, in xx seconds, log file, local ip and bantype
#
# This script is distributed under GPL V2 ans thus can be modified and redistributed
# by following the terms and conditions that can be found here:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
##############################################################################################################
use strict;
use IO::Socket;
#use warnings;
 
# Main configuration values. Change these to your needs
my %conf = (attempts => 5,
            min_seconds => 3,
            max_seconds => 15,
            logfile => "/var/log/autoban/main.log",
            localip => "192.168.5.5",
            bantype => "hosts", #Correct values: iptables|hosts
	        listenip => "localhost", #Ip on which the script will listen
	        listenport => "10200", #Port on which the script will listen
	        maxclients => "1", #Max numbers of clients allowed to send datas. 0 for unlimited
	        cleanup_period => "week", #Correct values: day|week|month
	      );

# Global vars
my (%hash,@date_num);
my @error = ("Failed password","Invalid user","11: Bye Bye","Failed publickey");
my $lcount = 0;
my @ip_array;

# hosts.deny cleanup function.
my $cmdlinearg = undef;
($cmdlinearg) = @ARGV;
# Let's ignore anything else than "-c"
if ( @ARGV > 0 && $cmdlinearg eq "-c" ) {
    cleanup();
}

sub cleanup {
    my @banned_ips;
    my $date = `date -d "1 $conf{cleanup_period} ago" +%s`;
    chomp($date); 
    # Opening /etc/hosts.deny and its backup, copying it while listing lines to be removed
    open my $DENY, "< /etc/hosts.deny" or die "Cannot open /etc/hosts.deny for reading: $!";
    open my $DENYBKP, "> /tmp/hosts.deny.bkp" or die "Cannot open /tmp/hosts.deny.bkp for writing $!";
	while (defined(my $line = <$DENY>)) {
	    print $DENYBKP "$line";
	    date_match_and_delete($line,$date) if ($line =~ /#sshd_autoban \d*/);
	}
    close $DENY  or die "Cannot close /etc/hosts.deny: $!";
    close $DENYBKP or die "Cannot close /tmp/hosts.deny.bkp: $!";

    open my $FILE, ">> $conf{logfile}" or die "Cannot open $conf{logfile} for writing: $!";
    if ( $lcount > 0 ) {
	my $ndate = `date +'[%D %T]'`;
	chomp($ndate);
        print $FILE "$ndate Cleanup job > $lcount old ips found and deleted in hosts.deny:\n";
        foreach (@ip_array) {
	    print $FILE "- $_\n";
	}
    } else {
	my $ndate = `date +'[%D %T]'`;
        chomp($ndate);
        print $FILE "$ndate Cleanup job > $lcount old ips found in hosts.deny. No deletion\n";
    } 
    close $FILE or die "Cannot close $conf{logfile}: $!";

    exit;
}

sub date_match_and_delete {
    my ($line,$date) = @_;
    my @line_array = split(" ", $line);
    if ($line_array[3] <= $date) {
	$lcount++;
	push(@ip_array, $line_array[1]);
	# Easy way deletion
	`perl -pi -e 's/$line//' /etc/hosts.deny`;
    }
}
# End of cleanup function


# Start listening on socket and waiting for syslog datas
my $sock = new IO::Socket::INET(
    LocalHost => $conf{listenip},
    LocalPort => $conf{listenport},
    Proto => "tcp",
    Listen => $conf{maxclients}) or die "Unable to create socket: $!";

my ($new_sock, $c_addr);

while (($new_sock, $c_addr) = $sock->accept()) {
    my ($client_port, $c_ip) = sockaddr_in($c_addr);
    my $client_ipnum = inet_ntoa($c_ip);

    open my $FILE, ">> $conf{logfile}" or die "Cannot open $conf{logfile} for writing: $!";
	my $ndate = `date +'[%D %T]'`;
	chomp($ndate);
	print $FILE "$ndate *** Starting logs analysis from: [$client_ipnum]\n";
    close $FILE or die "Cannot close $conf{logfile}: $!";

    while (defined ( my $line = <$new_sock>)) {
	chomp($line);
	#Ignore non erroneous connections and process lines that contains errors
	foreach my $error_ (@error) {
	    get_ip($line) if($line =~ /$error_/);
	}
    }
}

sub get_ip {
    my ($data) = @_;
    my $ip = $1 if($data =~ /(([\d]+\.){3}[\d]+)/);
 
    my $date = `date +%s`;
    chomp($date);
 
    check_it($date,$ip);
 
}
 
sub check_it {
    my ($date,$ip) = @_;
 
    if (defined ($hash{$ip})) {
        @date_num = @{$hash{$ip}};
        $date_num[1]++;
 
        if ( $date_num[1] >= $conf{attempts} ) {
            my $time = $date - $date_num[0];
            ban_ip($ip,$time,$date_num[1]) if( ($time > $conf{min_seconds}) && ($time <= $conf{max_seconds}) );
            # Prevent false positive
            delete($hash{$ip}) if($time > $conf{max_seconds});
        }
    }else {
        @date_num = ($date,'1');
        $hash{$ip} = \@date_num;
    }
 
}
 
sub ban_ip {
    my ($ip,$time,$num) = @_;
    my $ndate = `date +'[%D %T]'`;
    my $sdate = `date +%s`;
    chomp($ndate); 
    chomp($sdate);
 
    # Forbid every access from his/her ip
    # Check the way to proceed:
    if ( $conf{bantype} eq "iptables" ) {
        system "iptables -I INPUT -s $ip -d $conf{localip} -j DROP";
    } elsif ( $conf{bantype} eq "hosts" ) {
        open my $DENY, ">> /etc/hosts.deny" or die "Cannot open /etc/hosts.deny for writing: $!";
	# Writing deny informations plus an sshd autoban flag 
        print $DENY "ALL: $ip #sshd_autoban $sdate\n";
        close $DENY;
    }else {
        die "Wrong bantype option!";
    }
 
    # Log ban
    open my $FILE, ">> $conf{logfile}" or die "Cannot open $conf{logfile} for writing: $!";
    print $FILE "$ndate $ip Banned: $num invalid attempts in $time"."s\n";
    close $FILE or die "Cannot close $conf{logfile}: $!";
 
    # Delete ip from hash, we shouldn't see it again :)
    delete($hash{$ip});
 
}
